#include "plugin.hpp"
#include <htmlhelp.h>
#include <values.h> //MAXINT
#include <string.h>
#include "common.hpp"

static struct PluginStartupInfo Info;
static FARSTANDARDFUNCTIONS FSF;

typedef HWND (WINAPI *FPHH) (HWND, LPCSTR, UINT, DWORD);

enum
{
  MTitle,
  MMenuCommands,
  MDesc,
  MFile,
  MMask,
  MCurrentHelp,
  MLastHelp,
  MActivate,
  MShowInPanels,
  MShowInDisks,
  MDisksHotkey,
  MDeleteAsk,
  MMueller,
  MGeneratingHash,
  MErrTitle,
  MErrEditor,
  MErrX,
  MOk,
  MCancel,
};

static FPHH htmlHelp;

static BOOL IsOldFAR, NoHH;

// Plugin Registry root
static char PluginRootKey[80];

// This chars aren't letters
static char WordDiv[80];
static int WordDivLen;

static int ShowInPanels,
           ShowInDisks,
           DisksHotkey;

BOOL MyIsAlpha(unsigned char c);
BOOL FindBounds(const char *Str, int Len, int Pos, int *Start, int *End);
int FindEnd(const char *Str, int Len, int Pos);
int FindStart(const char *Str, int Len, int Pos);
int DoMain(char *FileName, char *DefKeyword, int OpenFrom);

#if defined(__BORLANDC__)
extern "C"
{
  BOOL WINAPI _export LibMain(HINSTANCE hDLLInst, DWORD fdwReason, LPVOID lpvReserved)
  {
    (void) hDLLInst;
    (void) fdwReason;
    (void) lpvReserved;
    return TRUE;
  }
}
#endif

const char *GetMsg(int MsgId) { return(Info.GetMsg(Info.ModuleNumber,MsgId)); }

static char *strstr(const char * str1, const char * str2 )
{
  char *cp = (char *) str1;
  char *s1, *s2;

  if ( !*str2 )
    return((char *)str1);

  while (*cp)
  {
    s1 = cp;
    s2 = (char *) str2;

    while ( *s1 && *s2 && !(*s1-*s2) )
      s1++, s2++;

    if (!*s2)
      return(cp);

    cp++;
  }

  return(NULL);
}

static void *memchr (const void * buf, int chr, size_t cnt)
{
  while ( cnt && (*(unsigned char *)buf != (unsigned char)chr) )
  {
    buf = (unsigned char *)buf + 1;
    cnt--;
  }
  return(cnt ? (void *)buf : NULL);
}

static int SayError(const char *s)
{
  const char *err[3];
  err[0] = GetMsg(MErrTitle);
  err[1] = s;
  err[2] = GetMsg(MOk);
  Info.Message( Info.ModuleNumber, FMSG_WARNING, NULL, err, 3, 1 );
  return 0;
}

//define USE_MALLOC
#include "memory.cpp"
#include "dialoginit.cpp"
#include "mueller.cpp"

static HKEY hKey;
static char FullKeyName[512];

HKEY CreateRegKey(HKEY hRoot,const char *Key)
{
  DWORD Disposition;
  FSF.sprintf(FullKeyName,"%s",PluginRootKey);
  if (*Key)
  {
    lstrcat(FullKeyName,"\\");
    lstrcat(FullKeyName,Key);
  }
  RegCreateKeyEx(hRoot,FullKeyName,0,NULL,0,KEY_WRITE,NULL,
                 &hKey,&Disposition);
  return(hKey);
}

HKEY OpenRegKey(HKEY hRoot,const char *Key)
{
  FSF.sprintf(FullKeyName,"%s",PluginRootKey);
  if (*Key)
  {
    lstrcat(FullKeyName,"\\");
    lstrcat(FullKeyName,Key);
  }
  if (RegOpenKeyEx(hRoot,FullKeyName,0,KEY_QUERY_VALUE,&hKey)!=ERROR_SUCCESS)
    return(NULL);
  return(hKey);
}

int GetRegKey(HKEY hRoot,const char *Key,const char *ValueName,char *ValueData,char *Default,DWORD DataSize)
{
  DWORD Type;
  int ExitCode;
  hKey=OpenRegKey(hRoot,Key);
  ExitCode=RegQueryValueEx(hKey,ValueName,0,&Type,(LPBYTE)ValueData,&DataSize);
  RegCloseKey(hKey);
  if (hKey==NULL || ExitCode!=ERROR_SUCCESS)
  {
    FSF.sprintf(ValueData,"%s",Default);
    return(FALSE);
  }
  return(TRUE);
}

void SetRegKey(HKEY hRoot,const char *Key,const char *ValueName,char *ValueData)
{
  hKey=CreateRegKey(hRoot,Key);
  RegSetValueEx(hKey,ValueName,0,REG_SZ,(CONST BYTE *)ValueData,lstrlen(ValueData)+1);
  RegCloseKey(hKey);
}

void DelRegValue(HKEY hRoot, char* ValueName)
{
  FSF.sprintf(FullKeyName,"%s",PluginRootKey);
  if (RegOpenKeyEx(hRoot,FullKeyName,0,KEY_WRITE,&hKey)==ERROR_SUCCESS)
  {
    RegDeleteValue(hKey,ValueName);
    RegCloseKey(hKey);
  }
}

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *psi)
{
  char AddWordDiv[sizeof(WordDiv)];
  HINSTANCE inst;

  Info=*psi;
  IsOldFAR=TRUE;
  if (psi->StructSize >= sizeof(struct PluginStartupInfo))
  {
    FSF=*psi->FSF;
    Info.FSF=&FSF;
    IsOldFAR=FALSE;
    //#ifndef USE_MALLOC
    //InitHeap();
    //#endif
    NoHH = FALSE;
    FSF.sprintf(PluginRootKey,"%s\\Active-Help",Info.RootKey);

    GetRegKey(HKEY_CURRENT_USER,"","ShowInPanels",AddWordDiv,"1",sizeof(AddWordDiv));
    FSF.sscanf(AddWordDiv,"%i",&ShowInPanels);
    GetRegKey(HKEY_CURRENT_USER,"","ShowInDisks",AddWordDiv,"0",sizeof(AddWordDiv));
    FSF.sscanf(AddWordDiv,"%i",&ShowInDisks);
    GetRegKey(HKEY_CURRENT_USER,"","DisksHotkey",AddWordDiv,"0",sizeof(AddWordDiv));
    FSF.sscanf(AddWordDiv,"%i",&DisksHotkey);
    if (DisksHotkey < 0 || DisksHotkey > 9)
      DisksHotkey = 0;

    WordDivLen = Info.AdvControl(Info.ModuleNumber, ACTL_GETSYSWORDDIV, WordDiv);
    //GetRegKey(HKEY_CURRENT_USER,"","AddWordDiv",AddWordDiv,"#",sizeof(AddWordDiv));
    //WordDivLen += lstrlen(AddWordDiv);
    //lstrcat(WordDiv, AddWordDiv);
    WordDivLen += sizeof(" \n\r\t");
    lstrcat(WordDiv, " \n\r\t");

    inst = LoadLibrary("HHCTRL.OCX");
    if (inst)
    {
      htmlHelp = (FPHH) GetProcAddress(inst,"HtmlHelpA");
      if (!htmlHelp)
        NoHH = TRUE;
    }
    else
    {
      NoHH = TRUE;
    }
  }
}

void WINAPI _export GetPluginInfo(struct PluginInfo *pi)
{
  static const char *MenuStrings[1];

  if(!IsOldFAR)
  {
    pi->StructSize=sizeof(struct PluginInfo);
    pi->Flags=PF_EDITOR;
    if (!ShowInPanels)
      pi->Flags|=PF_DISABLEPANELS;
    MenuStrings[0]=GetMsg(MTitle);
    pi->PluginMenuStrings=MenuStrings;
    pi->PluginMenuStringsNumber=1;
    pi->PluginConfigStrings=MenuStrings;
    pi->PluginConfigStringsNumber=1;
    if (ShowInDisks)
    {
      pi->DiskMenuStrings=MenuStrings;
      pi->DiskMenuStringsNumber=1;
      pi->DiskMenuNumbers = &DisksHotkey;
    }
  }
}

int WINAPI _export GetMinFarVersion(void)
{
  return MAKEFARVERSION(1,70,1282);
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom,int Item)
{
  int i,j,ret;
  struct EditorConvertText ect;
  struct EditorGetString egs;
  char *Keyword;
  char *LastKeyword;
  char FileName[512];
  char DefKeyword[512];
  struct EditorInfo ei;
  int Start, End;
  HH_AKLINK link;
  char tmp[10];

  if (IsOldFAR)
    return INVALID_HANDLE_VALUE;

  if (OpenFrom == OPEN_EDITOR)
  {
    Info.EditorControl(ECTL_GETINFO,(void *)&ei);
    FSF.sprintf(FileName,"%s",ei.FileName);
  }
  *DefKeyword=0;
  ret = DoMain(FileName,DefKeyword,OpenFrom);

  if (ret == -1 || (ret == 2 && OpenFrom != OPEN_EDITOR))
    return INVALID_HANDLE_VALUE;

  if (ret == 0)
  {
    egs.StringNumber=-1;
    if (Info.EditorControl(ECTL_GETSTRING,(void *)&egs))
    {
      if (egs.SelStart != -1 && egs.SelStart != egs.SelEnd && egs.SelStart != egs.StringLength)
      {
        i = 1;
        Start = egs.SelStart;
        End = (egs.SelEnd == -1 ? egs.StringLength : egs.SelEnd);
      }
      else
      {
        Start = 0;
        End = egs.StringLength;
        i = FindBounds(egs.StringText,egs.StringLength,ei.CurPos,&Start,&End);
      }
      if (i)
      {
        Keyword = (char *)malloc(End-Start+1);
        for (i=0, j=Start; j<End; i++, j++)
          Keyword[i] = egs.StringText[j];
        Keyword[i] = '\0';
        ect.Text = Keyword;
        ect.TextLength = End-Start;
        Info.EditorControl(ECTL_EDITORTOOEM,(void *)&ect);
        OemToCharBuff(Keyword,Keyword,End-Start);
      }
      else
        ret = 4;//return INVALID_HANDLE_VALUE;
    }
    else
    {
      SayError(GetMsg(MErrEditor));
      return INVALID_HANDLE_VALUE;
    }
  }

  if (ret == 1)
  {
    GetRegKey(HKEY_CURRENT_USER,"","LastKeywordLength",tmp,"0",sizeof(tmp));
    FSF.sscanf(tmp,"%i",&j);
    LastKeyword = (char *)malloc((j+1)*sizeof(char));
    GetRegKey(HKEY_CURRENT_USER,"","LastKeyword",LastKeyword,"",j+1);
    GetRegKey(HKEY_CURRENT_USER,"","LastHelpFile",FileName,"",sizeof(FileName));
    if (LastKeyword[0] && FileName[0])
      Keyword = LastKeyword;
    else
      return INVALID_HANDLE_VALUE;
  }

  int OK = TRUE;

  if (ret == 0 || ret == 1 || *(DefKeyword))
  {
    if ( ret > 1)
    {
      Keyword = DefKeyword;
      ret = 0;
    }
    if (!FSF.LStrnicmp(FileName,"hlf://",lstrlen("hlf://")))
    {
      OK = Info.ShowHelp(FileName+lstrlen("hlf://"),DefKeyword,FHELP_CUSTOMFILE|FHELP_USECONTENTS);
    }
    else if (!FSF.LStrnicmp(FileName,"mueller://",lstrlen("mueller://")))
    {
      OK = ShowMueller(FileName+lstrlen("mueller://"),Keyword);
    }
    else
    {
      if (FSF.LStricmp(&FileName[lstrlen(FileName)-4],".hlp") && !NoHH)
      {
        link.cbStruct = sizeof(HH_AKLINK);
        link.fReserved = FALSE;
        link.pszKeywords = Keyword;
        link.pszUrl = NULL;
        link.pszMsgText = NULL;
        link.pszMsgTitle = NULL;
        link.pszWindow = NULL;
        link.fIndexOnFail = TRUE;
        HWND hwnd = htmlHelp(GetDesktopWindow(), FileName, HH_KEYWORD_LOOKUP, (DWORD)&link);
        OK = hwnd==NULL ? FALSE : TRUE;
        ShowWindow(hwnd, SW_SHOWNORMAL);
      }
      else
      {
        OK = WinHelp(GetDesktopWindow(), FileName, HELP_KEY, (DWORD)Keyword);
      }
    }
    if (Keyword)
    {
      SetRegKey(HKEY_CURRENT_USER,"","LastKeywordLength",FSF.itoa(lstrlen(Keyword),tmp,10));
      SetRegKey(HKEY_CURRENT_USER,"","LastKeyword",Keyword);
      SetRegKey(HKEY_CURRENT_USER,"","LastHelpFile",FileName);
      if (ret <=1)
        free(Keyword);
    }
  }

  if (ret == 4)
  {
    if (!FSF.LStrnicmp(FileName,"hlf://",lstrlen("hlf://")))
    {
      OK = Info.ShowHelp(FileName+lstrlen("hlf://"),NULL,FHELP_CUSTOMFILE|FHELP_USECONTENTS);
    }
    else if (!FSF.LStrnicmp(FileName,"mueller://",lstrlen("mueller://")))
    {
      OK = ShowMueller(FileName+lstrlen("mueller://"),"");
    }
    else
    {
      if (FSF.LStricmp(&FileName[lstrlen(FileName)-4],".hlp") && !NoHH)
      {
        HWND hwnd = htmlHelp(GetDesktopWindow(), FileName, HH_HELP_FINDER, NULL);
        OK = hwnd==NULL ? FALSE : TRUE;
        ShowWindow(hwnd, SW_SHOWNORMAL);
      }
      else
      {
        OK = WinHelp(GetDesktopWindow(), FileName, HELP_FINDER, 0);
      }
    }
  }

  if (ret == 2)
  {
    egs.StringNumber=-1;
    if (Info.EditorControl(ECTL_GETSTRING,(void *)&egs))
    {
      if (egs.SelStart != -1 && egs.SelStart != egs.SelEnd && egs.SelStart != egs.StringLength)
      {
        Start = egs.SelStart;
        End = (egs.SelEnd == -1 ? egs.StringLength : egs.SelEnd);
      }
      else
      {
        if (ei.CurPos >= egs.StringLength || !egs.StringLength)
          return INVALID_HANDLE_VALUE;
        Start = ei.CurPos;
        while (egs.StringText[Start] != 9 && egs.StringText[Start] != 32 && Start >= 0)
          Start--;
        End = ei.CurPos;
        while (egs.StringText[End] != 9 && egs.StringText[End] != 32 && End < egs.StringLength)
          End++;
        if (End == Start)
          return INVALID_HANDLE_VALUE;
        Start++;
      }
      Keyword = (char *)malloc(End-Start+1);
      for (i=0, j=Start; j<End; i++, j++)
        Keyword[i] = egs.StringText[j];
      Keyword[i] = '\0';
      ect.Text = Keyword;
      ect.TextLength = End-Start;
      Info.EditorControl(ECTL_EDITORTOOEM,(void *)&ect);
      OemToCharBuff(Keyword,Keyword,End-Start);
    }
    else
      return INVALID_HANDLE_VALUE;
    ShellExecute(GetDesktopWindow(), "open", Keyword, NULL, NULL, SW_SHOWNORMAL);
    if (Keyword)
      free(Keyword);
  }
  if (!OK)
  {
    SayError(GetMsg(MErrX));
  }
  return INVALID_HANDLE_VALUE;
}

static char Value[50];

int GetHelpInfo(char *fd, char *fn, char *fm, int i)
{
  FSF.sprintf(Value,"File%d",i);
  if (!GetRegKey(HKEY_CURRENT_USER,"",Value,fn,"",512))
    return FALSE;
  FSF.sprintf(Value,"Desc%d",i);
  if (!GetRegKey(HKEY_CURRENT_USER,"",Value,fd,"",512))
    return FALSE;
  FSF.sprintf(Value,"Mask%d",i);
  if (!GetRegKey(HKEY_CURRENT_USER,"",Value,fm,"",512))
    return FALSE;
  return TRUE;
}

void SetHelpInfo(char *fd, char *fn, char *fm, int i)
{
  FSF.sprintf(Value,"Desc%d",i);
  SetRegKey(HKEY_CURRENT_USER,"",Value,fd);
  FSF.sprintf(Value,"File%d",i);
  SetRegKey(HKEY_CURRENT_USER,"",Value,fn);
  FSF.sprintf(Value,"Mask%d",i);
  SetRegKey(HKEY_CURRENT_USER,"",Value,fm);
}

void DelHelpInfo(int i)
{
  FSF.sprintf(Value,"Desc%d",i);
  DelRegValue(HKEY_CURRENT_USER,Value);
  FSF.sprintf(Value,"File%d",i);
  DelRegValue(HKEY_CURRENT_USER,Value);
  FSF.sprintf(Value,"Mask%d",i);
  DelRegValue(HKEY_CURRENT_USER,Value);
}

int DoMain(char *FileName, char *DefKeyword, int OpenFrom)
{
  #define MAX_ITEMS ('Z' - 'A' + 1)
  #define MENU_ITEMS (MAX_ITEMS + 1 + 4)
  char *HelpFileNames[MAX_ITEMS];
  char *Mask[MAX_ITEMS];
  char *tmp;
  struct FarMenuItem MenuItems[MENU_ITEMS], mi_tmp;
  char fn[512], fd[512], fm[512];
  int i, j, TotalItems, SelectedItem, LastPos, x;
  int ret;
  char szPath[512];
  HKEY hKeyCol, hKeyLang, hKeyLoc;
  DWORD dwSize, dwType;
  int BreakCode;
  static const int BreakKeys[] =
  {
    VK_INSERT,
    VK_DELETE,
    VK_F4,
    (PKF_CONTROL<<16)|VK_DOWN,
    (PKF_CONTROL<<16)|VK_UP,
    VK_SPACE,
    VK_RETURN,
    0
  };
  static const char DescHistory[] = "Active-Help-Desc";
  static const char PathHistory[] = "Active-Help-Path";
  static const char MaskHistory[] = "Active-Help-Mask";
  #define DLG_WIDTH 77
  #define DLG_HEIGHT 12
  static const struct InitDialogItem PreDialogItems[] =
  {
    DI_DOUBLEBOX   ,3  ,1  ,73 ,10 ,0                  ,MTitle,
    DI_TEXT        ,5  ,2  ,0  ,0  ,0                  ,MDesc,
    DI_EDIT        ,5  ,3  ,70 ,0  ,DIF_HISTORY        ,-1,
    DI_TEXT        ,5  ,4  ,0  ,0  ,0                  ,MFile,
    DI_EDIT        ,5  ,5  ,70 ,0  ,DIF_HISTORY        ,-1,
    DI_TEXT        ,5  ,6  ,0  ,0  ,0                  ,MMask,
    DI_EDIT        ,5  ,7  ,70 ,0  ,DIF_HISTORY        ,-1,
    DI_TEXT        ,0  ,8  ,0  ,0  ,DIF_SEPARATOR      ,-1,
    DI_BUTTON      ,0  ,9  ,0  ,0  ,DIF_CENTERGROUP    ,MOk,
    DI_BUTTON      ,0  ,9  ,0  ,0  ,DIF_CENTERGROUP    ,MCancel
  };
  struct FarDialogItem DialogItems[sizeof(PreDialogItems)/sizeof(PreDialogItems[0])];
  static const char *msg[2];

  InitDialogItems(PreDialogItems,DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
  DialogItems[2].Focus = TRUE;
  DialogItems[2].History = DescHistory;
  DialogItems[4].History = PathHistory;
  DialogItems[6].History = MaskHistory;
  DialogItems[8].DefaultButton = 1;

  msg[0] = GetMsg(MTitle);
  msg[1] = GetMsg(MDeleteAsk);

  for (i=0; i<4; i++)
    MenuItems[i].Checked=MenuItems[i].Selected=MenuItems[i].Separator=0;

  FSF.sprintf(MenuItems[0].Text,"%s",GetMsg(MCurrentHelp));
  FSF.sprintf(MenuItems[1].Text,"%s",GetMsg(MLastHelp));
  FSF.sprintf(MenuItems[2].Text,"%s",GetMsg(MActivate));
  MenuItems[3].Separator=1;

  j = -1;
  for (i=0; i<MAX_ITEMS; i++)
  {
    if (!GetHelpInfo(fd,fn,fm,i))
      break;
    Mask[i]=(char *)malloc(sizeof(char)*(lstrlen(fm)+1));
    HelpFileNames[i]=(char *)malloc(sizeof(char)*(lstrlen(fn)+1));
    FSF.sprintf(Mask[i],"%s",fm);
    FSF.sprintf(HelpFileNames[i],"%s",fn);
    FSF.sprintf(MenuItems[i+4].Text,"&%c. %.*s",i+'A',66,fd);
    MenuItems[i+4].Checked=MenuItems[i+4].Selected=MenuItems[i+4].Separator=0;
    if (OpenFrom == OPEN_EDITOR && j == -1 && fm[0] && FSF.ProcessName(fm,FileName,PN_CMPNAMELIST|PN_SKIPPATH))
      j = i;
  }
  FSF.sprintf(MenuItems[i+4].Text,"%*c",66/2,' ');
  MenuItems[i+4].Checked=MenuItems[i+4].Selected=MenuItems[i+4].Separator=0;
  TotalItems = i;
  SelectedItem = 0;
  if (j != -1)
    SelectedItem = j+4;
  else
  {
    if (TotalItems)
    {
      GetRegKey(HKEY_CURRENT_USER,"","SelectedItem",Value,"4",sizeof(Value));
      FSF.sscanf(Value,"%i",&SelectedItem);
      if (SelectedItem < 4 || SelectedItem > (TotalItems+4-1))
        SelectedItem = 4;
    }
  }
  GetRegKey(HKEY_CURRENT_USER,"","LastPos",Value,"0",sizeof(Value));
  FSF.sscanf(Value,"%i",&LastPos);
  if (LastPos < 0 || LastPos > (TotalItems+4))
    LastPos = 0;

  *(DefKeyword) = 0;
  ret = -1;
  while (1)
  {
    if (ret != -1)
      break;
    if (SelectedItem)
      MenuItems[SelectedItem].Checked = '*';
    MenuItems[LastPos].Selected = 1;

    i = Info.Menu(Info.ModuleNumber,-1,-1,0,
                  FMENU_WRAPMODE|FMENU_AUTOHIGHLIGHT,
                  GetMsg(MTitle),GetMsg(MMenuCommands),"Usage",BreakKeys,&BreakCode,
                  MenuItems,TotalItems+1+4);

    if (i>=0)
    {
      MenuItems[LastPos].Selected = 0;
      LastPos = i;
      if (BreakCode == -1 && i >= 4 && TotalItems)
        BreakCode = 5;
      switch (BreakCode)
      {
        case 0:
          DialogItems[2].Focus = TRUE;
          DialogItems[2].Data[0] = 0;
          DialogItems[4].Data[0] = 0;
          DialogItems[6].Data[0] = 0;
          i = Info.Dialog(Info.ModuleNumber,-1,-1,DLG_WIDTH,DLG_HEIGHT,"Settings",(struct FarDialogItem *)&DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
          if (i!=8 || !DialogItems[4].Data[0] || !DialogItems[2].Data[0] || TotalItems >= MAX_ITEMS)
            break;
          if (!TotalItems)
            SelectedItem = 4;
          LastPos = TotalItems+4;
          TotalItems++;
          HelpFileNames[TotalItems-1]=(char *)malloc(sizeof(char)*(lstrlen(DialogItems[4].Data)+1));
          Mask[TotalItems-1]=(char *)malloc(sizeof(char)*(lstrlen(DialogItems[6].Data)+1));
          FSF.sprintf(Mask[TotalItems-1],"%s",DialogItems[6].Data);
          FSF.sprintf(HelpFileNames[TotalItems-1],"%s",DialogItems[4].Data);
          FSF.sprintf(MenuItems[TotalItems-1+4].Text,"&%c. %.*s",TotalItems+'A'-1,66,DialogItems[2].Data);
          MenuItems[TotalItems-1+4].Checked=MenuItems[TotalItems-1+4].Selected=MenuItems[TotalItems-1+4].Separator=0;
          MenuItems[TotalItems+4].Checked=MenuItems[TotalItems+4].Selected=MenuItems[TotalItems+4].Separator=0;
          FSF.sprintf(MenuItems[TotalItems+4].Text,"%*c",66/2,' ');
          SetHelpInfo(&MenuItems[TotalItems-1+4].Text[4],HelpFileNames[TotalItems-1],Mask[TotalItems-1],TotalItems-1);
          break;

        case 1:
          if (i >= TotalItems+4 || i < 4 || !TotalItems)
            break;
          if (Info.Message(Info.ModuleNumber,FMSG_MB_YESNO,NULL,msg,2,0) != 0)
            break;
          if (i == TotalItems+4-1 && TotalItems > 1)
            LastPos--;
          if (SelectedItem >= i)
          {
            MenuItems[SelectedItem].Checked = 0;
            SelectedItem = SelectedItem==i ? 4 : SelectedItem-1;
            if (TotalItems == 1)
              SelectedItem = 0;
          }
          for (i=i-4; i<TotalItems-1; i++)
          {
            MenuItems[i+4] = MenuItems[i+1+4];
            free(HelpFileNames[i]);
            HelpFileNames[i] = HelpFileNames[i+1];
            HelpFileNames[i+1] = NULL;
            free(Mask[i]);
            Mask[i] = Mask[i+1];
            Mask[i+1] = NULL;
            SetHelpInfo(&MenuItems[i+4].Text[4],HelpFileNames[i],Mask[i],i);
          }
          MenuItems[i+4] = MenuItems[i+1+4];
          TotalItems--;
          DelHelpInfo(TotalItems);
          for (i=0; i<TotalItems; i++)
            MenuItems[i+4].Text[1] = i + 'A';
          break;

        case 2:
          if (i>=TotalItems+4 || i < 4)
            break;
          i = i - 4;
          FSF.sprintf(DialogItems[2].Data,"%s",&MenuItems[i+4].Text[4]);
          FSF.sprintf(DialogItems[4].Data,"%s",HelpFileNames[i]);
          FSF.sprintf(DialogItems[6].Data,"%s",Mask[i]);
          DialogItems[2].Focus = TRUE;
          j = Info.Dialog(Info.ModuleNumber,-1,-1,DLG_WIDTH,DLG_HEIGHT,"Settings",(struct FarDialogItem *)&DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
          if (j!=8)
            break;
          free(HelpFileNames[i]);
          HelpFileNames[i]=(char *)malloc(sizeof(char)*(lstrlen(DialogItems[4].Data)+1));
          FSF.sprintf(HelpFileNames[i],"%s",DialogItems[4].Data);
          free(Mask[i]);
          Mask[i]=(char *)malloc(sizeof(char)*(lstrlen(DialogItems[6].Data)+1));
          FSF.sprintf(Mask[i],"%s",DialogItems[6].Data);
          FSF.sprintf(MenuItems[i+4].Text,"&%c. %.*s",i+'A',66,DialogItems[2].Data);
          SetHelpInfo(&MenuItems[i+4].Text[4],HelpFileNames[i],Mask[i],i);
          break;

        case 3:
          x = 1;
        case 4:
          if (BreakCode == 4)
            x = -1;
          if (BreakCode == 3 && (TotalItems == 1 || i >= TotalItems+4-1 || i < 4))
            break;
          if (BreakCode == 4 && (TotalItems == 1 || i >= TotalItems+4 || i <= 4))
            break;
          if (SelectedItem == i)
          {
            MenuItems[SelectedItem].Checked = 0;
            SelectedItem += x;
          }
          else
          {
            if (SelectedItem == i+x)
            {
              MenuItems[SelectedItem].Checked = 0;
              SelectedItem -= x;
            }
          }
          LastPos += x;
          i = i - 4;
          tmp = HelpFileNames[i+x];
          mi_tmp = MenuItems[i+x+4];
          HelpFileNames[i+x] = HelpFileNames[i];
          MenuItems[i+x+4] = MenuItems[i+4];
          MenuItems[i+4] = mi_tmp;
          HelpFileNames[i] = tmp;
          tmp = Mask[i+x];
          Mask[i+x] = Mask[i];
          Mask[i] = tmp;
          j = MenuItems[i+x+4].Text[1];
          MenuItems[i+x+4].Text[1] = (char)(j+x);
          MenuItems[i+4].Text[1] = (char)j;
          SetHelpInfo(&MenuItems[i+4].Text[4],HelpFileNames[i],Mask[i],i);
          SetHelpInfo(&MenuItems[i+x+4].Text[4],HelpFileNames[i+x],Mask[i+x],i+x);
          break;

        case 5:
          if (TotalItems && i >=4 && i < (TotalItems+4))
          {
            if (SelectedItem)
              MenuItems[SelectedItem].Checked = 0;
            SelectedItem = i;
          }
          break;

        default:
          if (TotalItems && i < (TotalItems+4))
          {
            if (i >=4)
            {
              if (SelectedItem)
                MenuItems[SelectedItem].Checked = 0;
              SelectedItem = i;
            }
            FSF.sprintf(FileName,"%s",HelpFileNames[SelectedItem-4]);
            if (!lstrcmp(FileName,"MSDN"))
            {
              if( RegOpenKeyEx(HKEY_LOCAL_MACHINE,"SOFTWARE\\Microsoft\\HTML Help Collections\\Developer Collections",0,KEY_READ,&hKeyCol)==ERROR_SUCCESS ){
                dwSize=sizeof(szPath);
                if( RegQueryValueEx(hKeyCol,"Language",0,&dwType,(LPBYTE)szPath,&dwSize)==ERROR_SUCCESS ){
                  dwSize=sizeof(szPath);
                  if( RegOpenKeyEx(hKeyCol,szPath,0,KEY_READ,&hKeyLang)==ERROR_SUCCESS ){
                    dwSize=sizeof(szPath);
                    if( RegQueryValueEx(hKeyLang,"Preferred",0,&dwType,(LPBYTE)szPath,&dwSize)==ERROR_SUCCESS ){
                      if( RegOpenKeyEx(hKeyLang,szPath,0,KEY_READ,&hKeyLoc)==ERROR_SUCCESS ){
                        dwSize=sizeof(szPath);
                        if( RegQueryValueEx(hKeyLoc,"Filename",0,&dwType,(LPBYTE)szPath,&dwSize)==ERROR_SUCCESS ){
                          FSF.sprintf(FileName,"%s",szPath);
                        }
                        RegCloseKey(hKeyLoc);
                      }
                    }
                    RegCloseKey(hKeyLang);
                  }
                }
                RegCloseKey(hKeyCol);
              }
            }
            if (strstr(FileName,"|"))
            {
              FSF.sprintf(DefKeyword,"%s",strstr(FileName,"|")+1);
              *(strstr(FileName,"|")) = 0;
            }
            ret = (i == 0 && OpenFrom != OPEN_EDITOR) ? 4 : (i < 4 ? i : 4);
          }
          break;
      }
      continue;
    }
    break;
  }

  for (i=0;i<TotalItems;i++)
  {
    free(HelpFileNames[i]);
    free(Mask[i]);
  }
  SetRegKey(HKEY_CURRENT_USER,"","SelectedItem",FSF.itoa(SelectedItem,Value,10));
  SetRegKey(HKEY_CURRENT_USER,"","LastPos",FSF.itoa(LastPos,Value,10));
  return ret;
}

// What we consider as letter
BOOL MyIsAlpha(unsigned char c)
{
  return ( memchr(WordDiv, c, WordDivLen)==NULL ? TRUE : FALSE );
}

// Finding word bounds (what'll be converted) (Str is in OEM)
BOOL FindBounds(const char *Str, int Len, int Pos, int *Start, int *End)
{
  int i = 1;
  BOOL ret = FALSE;
  int r = MAXINT;

  // If line isn't ""
  if( Len>*Start )
  {
    *End=min(*End,Len);

    // Pos between [Start, End] ?
    Pos=max(Pos,*Start);
    Pos=min(*End,Pos);

    // If current character is non letter
    if(!MyIsAlpha(Str[Pos]))
    {
      // Looking for letter on the left and counting radius
      while((*Start<=Pos-i) && (!MyIsAlpha(Str[Pos-i])))
        i++;

      // Letter was found on the left
      if(*Start<=Pos-i)
        r=i; // Storing radius

      i=1;
      // Looking for letter on the right and counting radius
      while((Pos+i<=*End) && (!MyIsAlpha(Str[Pos+i])))
        i++;

      // Letter was not found
      if(Pos+i>*End)
        i=MAXINT;

      // Here r is left radius and i is right radius

      // If no letters was found
      if( min(r,i)!=MAXINT )
      {
        // What radius is less? Left?
        if( r <= i )
        {
          *End=Pos-r+1;
          *Start=FindStart(Str, *Start, *End);
        }
        else // Right!
        {
          *Start=Pos+i;
          *End=FindEnd(Str, *Start, *End);
        }
        ret=TRUE;
      }
    }
    else // Current character is letter!
    {
      *Start=FindStart(Str, *Start, Pos);
      *End=FindEnd(Str, Pos, *End);
      ret=TRUE;
    }
  }

  if(!ret)
    *Start=*End=-1;

  return ret;
}

int FindStart(const char *Str, int Start, int End)
{
  // Current pos in Str
  int CurPos = End-1;

  // While current character is letter
  while( CurPos>=Start && MyIsAlpha(Str[CurPos]) )
    CurPos--; // Moving to left
  return CurPos+1;
}

int FindEnd(const char *Str, int CurPos, int End)
{
  // While current character is letter
  while( CurPos<End && MyIsAlpha(Str[CurPos]))
    CurPos++; // Moving to right
  return CurPos;
}

int WINAPI _export Configure(int ItemNumber)
{
  static const struct InitDialogItem PreDialogItems[] =
  {
    DI_DOUBLEBOX  ,3  ,1  ,60 ,8  ,0               ,MTitle,
    DI_CHECKBOX   ,5  ,2  ,0  ,0  ,0               ,MShowInPanels,
    DI_TEXT       ,4  ,3  ,0  ,0  ,DIF_SEPARATOR   ,-1,
    DI_CHECKBOX   ,5  ,4  ,0  ,0  ,0               ,MShowInDisks,
    DI_TEXT       ,9  ,5  ,0  ,0  ,0               ,MDisksHotkey,
    DI_FIXEDIT    ,7  ,5  ,7  ,0  ,DIF_MASKEDIT    ,-1,
    DI_TEXT       ,4  ,6  ,0  ,0  ,DIF_SEPARATOR   ,-1,
    DI_BUTTON     ,0  ,7  ,0  ,0  ,DIF_CENTERGROUP ,MOk,
    DI_BUTTON     ,0  ,7  ,0  ,0  ,DIF_CENTERGROUP ,MCancel
  };
  struct FarDialogItem DialogItems[sizeof(PreDialogItems)/sizeof(PreDialogItems[0])];
  char Value[10];
  static const char Mask[] = "9";

  InitDialogItems(PreDialogItems,DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
  DialogItems[1].Focus = TRUE;
  DialogItems[1].Selected = ShowInPanels;
  DialogItems[3].Selected = ShowInDisks;
  DialogItems[7].DefaultButton = 1;
  DialogItems[5].Mask = Mask;
  FSF.sprintf(DialogItems[5].Data,"%i",DisksHotkey);

  if (Info.Dialog(Info.ModuleNumber,-1,-1,64,10,"Configure",(struct FarDialogItem *)&DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0])) != 7)
    return FALSE;

  ShowInPanels = DialogItems[1].Selected;
  ShowInDisks = DialogItems[3].Selected;
  FSF.sscanf(DialogItems[5].Data,"%i",&DisksHotkey);

  SetRegKey(HKEY_CURRENT_USER,"","ShowInPanels",FSF.itoa(ShowInPanels,Value,10));
  SetRegKey(HKEY_CURRENT_USER,"","ShowInDisks",FSF.itoa(ShowInDisks,Value,10));
  SetRegKey(HKEY_CURRENT_USER,"","DisksHotkey",FSF.itoa(DisksHotkey,Value,10));
  return TRUE;
}
