#include <windows.h>
#include <stdio.h>
#define BUFFER_SIZE (10*1024+1)

int isdigit(int c)
{
  return (c>='0' && c<='9' ? 1 : 0);
}

int isroman(int c)
{
  return (c=='I' || c=='V' || c=='X' ? 1 : 0);
}

bool ReadLine(HANDLE file,char *buffer,size_t len)
{
  DWORD CurrPos=SetFilePointer(file,0,NULL,FILE_CURRENT),transferred;
  bool res=true;
  if(ReadFile(file,buffer,len-1,&transferred,NULL)&&transferred)
  {
    buffer[len-1]=0;
    DWORD len=0; char *ptr=buffer;
    for(size_t i=0;i<transferred;i++,len++,ptr++)
    {
      if(buffer[i]=='\n'||buffer[i]=='\r')
      {
        if((buffer[i+1]=='\n'||buffer[i+1]=='\r')&&buffer[i]!=buffer[i+1])
          len++;
        break;
      }
    }
    *ptr=0;
    SetFilePointer(file,CurrPos+len+1,NULL,FILE_BEGIN);
  }
  else
  {
    buffer[0]=0;
    res=false;
  }
  return res;
}

void Format(int len, const char *str, int indent, HANDLE fp)
{
  #define RLIMIT 76
  const char *ptr = str;
  const char *end = &str[len];
  DWORD t;

  if (indent < 0)
  {
    while (indent < 0)
    {
      WriteFile(fp," ",1,&t,NULL);
      indent++;
    }
  }
  while (ptr < end)
  {
    while ((ptr < end) && ((ptr-str) < RLIMIT))
      ptr++;
    if ((*ptr != 0x20) && (ptr < end))
    {
      while ((ptr > str) && (*ptr != 0x20))
        ptr--;
      if ((*ptr != 0x20) || (ptr == str))
      {
        ptr = ((end-str) > (RLIMIT-1) ? ptr+RLIMIT : end);
      }
    }
    char *text = (char *)malloc(ptr-str+3);
    FSF.sprintf(text,"%.*s\r\n",ptr-str,str);
    WriteFile(fp,text,lstrlen(text),&t,NULL);
    free(text);
    while ((ptr < end) && (*ptr == 0x20))
      ptr++;
    str = ptr;
  }
  if (indent > 0)
  {
    while (indent > 0)
    {
      WriteFile(fp," ",1,&t,NULL);
      indent--;
    }
  }
}

int ShowMueller(const char *FileName, const char *Keyword)
{
  static const char MuellerHistory[] = "Active-Help-Mueller";
  #define MDLG_WIDTH 30
  #define MDLG_HEIGHT 5
  const struct InitDialogItem PreDialogItems[] =
  {
    DI_DOUBLEBOX   ,3  ,1  ,26 ,3 ,0                              ,MMueller,
    DI_EDIT        ,5  ,2  ,24 ,0 ,DIF_HISTORY|DIF_USELASTHISTORY ,-1,
  };
  struct FarDialogItem DialogItems[sizeof(PreDialogItems)/sizeof(PreDialogItems[0])];

  InitDialogItems(PreDialogItems,DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
  DialogItems[1].Focus = TRUE;
  DialogItems[1].History = MuellerHistory;
  DialogItems[1].DefaultButton = 1;
  FSF.sprintf(DialogItems[1].Data,"%.*s",512,Keyword);
  int ret = Info.Dialog(Info.ModuleNumber,-1,-1,MDLG_WIDTH,MDLG_HEIGHT,NULL,(struct FarDialogItem *)&DialogItems,sizeof(PreDialogItems)/sizeof(PreDialogItems[0]));
  if ((ret != 1) || (!*DialogItems[1].Data))
    return 1;
  char keyword[512];
  FSF.sprintf(keyword,"%s",DialogItems[1].Data);
  if (lstrlen(keyword) < 2)
    lstrcat(keyword," ");

  HANDLE fpDic = CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  char HashFileName[MAX_PATH];
  FSF.sprintf(HashFileName,"%s.hsh",FileName);
  HANDLE fpHash = CreateFile(HashFileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);

  if (fpDic==INVALID_HANDLE_VALUE)
    return 0;
  if (fpHash==INVALID_HANDLE_VALUE)
  {
    fpHash = CreateFile(HashFileName,GENERIC_WRITE|GENERIC_READ,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);

    if (fpHash==INVALID_HANDLE_VALUE)
    {
      CloseHandle(fpDic);
      return 0;
    }

    #define START "'c"
    #define PRESTART "'b"
    char DicItem[BUFFER_SIZE];
    DWORD pos = 0;

    const char *Msg[2];
    Msg[0] = GetMsg(MMueller);
    Msg[1] = GetMsg(MGeneratingHash);
    HANDLE ss = Info.SaveScreen(0,0,-1,-1);
    Info.Message(Info.ModuleNumber,0,NULL,Msg,2,0);

    while (ReadLine(fpDic,DicItem,BUFFER_SIZE) && FSF.LStrnicmp(START,DicItem,2))
    {
      pos = SetFilePointer(fpDic,0,NULL,FILE_CURRENT);
    }
    if (pos==INVALID_SET_FILE_POINTER || !*DicItem || !*(DicItem+1))
    {
      Info.RestoreScreen(ss);
      CloseHandle(fpHash);
      CloseHandle(fpDic);
      return 0;
    }
    SetFilePointer(fpDic,pos,NULL,FILE_BEGIN);
    DWORD hash;
    char HashItem[20];
    FSF.sprintf(HashItem,"%s",PRESTART);
    while (ReadLine(fpDic,DicItem,BUFFER_SIZE))
    {
      hash = pos;
      pos = SetFilePointer(fpDic,0,NULL,FILE_CURRENT);
      if (!FSF.LStrnicmp(DicItem,HashItem,2) || lstrlen(DicItem)<2/* || DicItem[1]==0x20*/)
        continue;
      FSF.sprintf(HashItem,"%.2s%d\r\n",DicItem,hash);
      WriteFile(fpHash,HashItem,lstrlen(HashItem),&hash,NULL);
    }
    SetFilePointer(fpDic,0,NULL,FILE_BEGIN);
    SetFilePointer(fpHash,0,NULL,FILE_BEGIN);
    Info.RestoreScreen(ss);
  }

  char HashItem[20];

  while (ReadLine(fpHash,HashItem,20) && FSF.LStrnicmp(keyword,HashItem,2))
    ;

  if (!*HashItem || lstrlen(HashItem) < 3)
  {
    CloseHandle(fpHash);
    CloseHandle(fpDic);
    return 1;
  }

  if (SetFilePointer(fpDic,FSF.atoi(HashItem+2),NULL,FILE_BEGIN)==INVALID_SET_FILE_POINTER)
  {
    CloseHandle(fpHash);
    CloseHandle(fpDic);
    return 0;
  }

  char DicItem[BUFFER_SIZE];
  bool found=false;
  while (ReadLine(fpDic,DicItem,BUFFER_SIZE) && !FSF.LStrnicmp(keyword,DicItem,2))
  {
    const char *ptr = strstr((const char *)DicItem,"  ");
    int len = (ptr ? ((ptr-DicItem) < 2 ? 2 : ptr-DicItem ) : 0);
    if (lstrlen(keyword) > len || len == 0)
      continue;
    if (!FSF.LStrnicmp(DicItem,keyword,lstrlen(keyword)))
    {
      found=true;
      break;
    }
  }

  if (found)
  {
    char TempFileName[MAX_PATH];
    HANDLE fpTemp;
    if (CreateDirectory(FSF.MkTemp(TempFileName,"ACTH"),NULL))
    {
      lstrcat(TempFileName,"\\mueller.txt");
      fpTemp = CreateFile(TempFileName,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);
      if (fpTemp==INVALID_HANDLE_VALUE)
      {
        CloseHandle(fpHash);
        CloseHandle(fpDic);
        return 0;
      }
    }
    else
    {
      CloseHandle(fpHash);
      CloseHandle(fpDic);
      return 0;
    }
    const char *str = strstr((const char *)DicItem,"  ");
    size_t len = str-DicItem;
    Format(len,DicItem,0,fpTemp);
    const char *end = &DicItem[lstrlen(DicItem)];
    const char *tmp, *ptr;
    str += 2;
    ptr = str;
    while (ptr<end)
    {
      tmp = ptr;
      if (*ptr == '_' && (*(ptr-1) == 0x20))
      {
        while (isroman(*(++ptr)))
          ;
        if (*ptr == 0x20)
        {
          Format(tmp-str,str,2,fpTemp);
          str = tmp;
          ptr++;
        }
      }
      else if (isdigit(*ptr) && (*(ptr-1) == 0x20))
      {
        while (isdigit(*(++ptr)))
          ;
        if (*ptr == '>')
        {
          Format(tmp-str,str,0,fpTemp);
          str = tmp;
          ptr++;
        }
        else if (*ptr == '.')
        {
          Format(tmp-str,str,2,fpTemp);
          str = tmp;
          ptr++;
        }
      }
      else if ((*(ptr+1) == '>') && (*(ptr-1) == 0x20))
      {
        const char *srch = tmp;
        const char *save = tmp;
        while ((srch > str) && (*srch != ';'))
          srch--;
        if (*srch == ';' && ((srch+2) != save))
        {
          tmp = srch+1;
        }
        Format(tmp-str,str,0,fpTemp);
        if (*srch == ';' && ((srch+2) != save))
        {
          tmp = save;
          Format(tmp-srch-2,srch+2,-1,fpTemp);
        }
        str = tmp;
        ptr+=2;
      }
      else
      {
        ptr++;
      }
    }
    Format(lstrlen(str),str,0,fpTemp);
    CloseHandle(fpTemp);
    Info.Editor(TempFileName,keyword,0,0,-1,-1,EF_DELETEONCLOSE,0,1);
  }

  CloseHandle(fpHash);
  CloseHandle(fpDic);
  return 1;
}